#!/usr/bin/env python
# $URL$
# $Rev$
# Make ICC Profile

# References
#
# [ICC 2001] ICC Specification ICC.1:2001-04 (Profile version 2.4.0)
# [ICC 2004] ICC Specification ICC.1:2004-10 (Profile version 4.2.0.0)

import struct

# Local module.
import iccp

def black(m):
    """Return a function that maps all values from [0.0,m] to 0, and maps
    the range [m,1.0] into [0.0, 1.0] linearly.
    """

    m = float(m)

    def f(x):
        if x <= m:
            return 0.0
        return (x-m)/(1.0-m)
    return f

# For monochrome input the required tags are (See [ICC 2001] 6.3.1.1):
# profileDescription [ICC 2001] 6.4.32
# grayTRC [ICC 2001] 6.4.19
# mediaWhitePoint [ICC 2001] 6.4.25
# copyright [ICC 2001] 6.4.13

def agreyprofile(out):
    it = iccp.Profile().greyInput()
    it.addTags(kTRC=black(0.07))
    it.write(out)

def main():
    import sys
    agreyprofile(sys.stdout)

if __name__ == '__main__':
    main()



# This file comprises the tests that are internally validated (as
# opposed to tests which produce output files that are externally
# validated).  Primarily they are unittests.

# There is a read/write asymmetry: It is fairly easy to
# internally validate the results of reading a PNG file because we
# can know what pixels it should produce, but when writing a PNG
# file many choices are possible. The only thing we can do is read
# it back in again, which merely checks consistency, not that the
# PNG file we produce is valid.

# Run the tests from the command line:
#   python -c 'import test_png;test_png.runTest()'
# If you have nose installed you can use that:
#   nosetests .

# (For an in-memory binary file IO object) We use BytesIO where
# available, otherwise we use StringIO, but name it BytesIO.
try:
    from io import BytesIO
except ImportError:
    from StringIO import StringIO as BytesIO
import itertools
import struct
import sys
# http://www.python.org/doc/2.4.4/lib/module-unittest.html
import unittest
import zlib

from array import array

import png
import pngsuite

from png import strtobytes      # Don't do this at home.


def runTest():
    unittest.main(__name__)

def topngbytes(name, rows, x, y, **k):
    """
    Convenience function for creating a PNG file "in memory" as
    a string.  Creates a :class:`Writer` instance using the keyword
    arguments, then passes `rows` to its :meth:`Writer.write` method.
    The resulting PNG file is returned as a string.  `name` is used
    to identify the file for debugging.
    """

    import os

    print name
    f = BytesIO()
    w = png.Writer(x, y, **k)
    w.write(f, rows)
    if os.environ.get('PYPNG_TEST_TMP'):
        w = open(name, 'wb')
        w.write(f.getvalue())
        w.close()
    return f.getvalue()

def _redirect_io(inp, out, f):
    """Calls the function `f` with ``sys.stdin`` changed to `inp`
    and ``sys.stdout`` changed to `out`.  They are restored when `f`
    returns.  This function returns whatever `f` returns.
    """

    import os
    import sys

    oldin, sys.stdin = sys.stdin, inp
    oldout, sys.stdout = sys.stdout, out
    try:
        x = f()
    finally:
        sys.stdin = oldin
        sys.stdout = oldout
    if os.environ.get('PYPNG_TEST_TMP') and hasattr(out,'getvalue'):
        name = mycallersname()
        if name:
            w = open(name+'.png', 'wb')
            w.write(out.getvalue())
            w.close()
    return x

def mycallersname():
    """Returns the name of the caller of the caller of this function
    (hence the name of the caller of the function in which
    "mycallersname()" textually appears).  Returns None if this cannot
    be determined.
    """

    # http://docs.python.org/library/inspect.html#the-interpreter-stack
    import inspect

    frame = inspect.currentframe()
    if not frame:
        return None
    frame_,filename_,lineno_,funname,linelist_,listi_ = (
      inspect.getouterframes(frame)[2])
    return funname

def seqtobytes(s):
    """Convert a sequence of integers to a *bytes* instance.  Good for
    plastering over Python 2 / Python 3 cracks.
    """

    return strtobytes(''.join(chr(x) for x in s))

class Test(unittest.TestCase):
    # This member is used by the superclass.  If we don't define a new
    # class here then when we use self.assertRaises() and the PyPNG code
    # raises an assertion then we get no proper traceback.  I can't work
    # out why, but defining a new class here means we get a proper
    # traceback.
    class failureException(Exception):
        pass

    def helperLN(self, n):
        mask = (1 << n) - 1
        # Use small chunk_limit so that multiple chunk writing is
        # tested.  Making it a test for Issue 20 (googlecode).
        w = png.Writer(15, 17, greyscale=True, bitdepth=n, chunk_limit=99)
        f = BytesIO()
        w.write_array(f, array('B', map(mask.__and__, range(1, 256))))
        r = png.Reader(bytes=f.getvalue())
        x,y,pixels,meta = r.read()
        self.assertEqual(x, 15)
        self.assertEqual(y, 17)
        self.assertEqual(list(itertools.chain(*pixels)),
                         map(mask.__and__, range(1,256)))
    def testL8(self):
        return self.helperLN(8)
    def testL4(self):
        return self.helperLN(4)
    def testL2(self):
        "Also tests asRGB8."
        w = png.Writer(1, 4, greyscale=True, bitdepth=2)
        f = BytesIO()
        w.write_array(f, array('B', range(4)))
        r = png.Reader(bytes=f.getvalue())
        x,y,pixels,meta = r.asRGB8()
        self.assertEqual(x, 1)
        self.assertEqual(y, 4)
        for i,row in enumerate(pixels):
            self.assertEqual(len(row), 3)
            self.assertEqual(list(row), [0x55*i]*3)
    def testP2(self):
        "2-bit palette."
        a = (255,255,255)
        b = (200,120,120)
        c = (50,99,50)
        w = png.Writer(1, 4, bitdepth=2, palette=[a,b,c])
        f = BytesIO()
        w.write_array(f, array('B', (0,1,1,2)))
        r = png.Reader(bytes=f.getvalue())
        x,y,pixels,meta = r.asRGB8()
        self.assertEqual(x, 1)
        self.assertEqual(y, 4)
        self.assertEqual(map(list, pixels), map(list, [a, b, b, c]))
    def testPtrns(self):
        "Test colour type 3 and tRNS chunk (and 4-bit palette)."
        a = (50,99,50,50)
        b = (200,120,120,80)
        c = (255,255,255)
        d = (200,120,120)
        e = (50,99,50)
        w = png.Writer(3, 3, bitdepth=4, palette=[a,b,c,d,e])
        f = BytesIO()
        w.write_array(f, array('B', (4, 3, 2, 3, 2, 0, 2, 0, 1)))
        r = png.Reader(bytes=f.getvalue())
        x,y,pixels,meta = r.asRGBA8()
        self.assertEqual(x, 3)
        self.assertEqual(y, 3)
        c = c+(255,)
        d = d+(255,)
        e = e+(255,)
        boxed = [(e,d,c),(d,c,a),(c,a,b)]
        flat = map(lambda row: itertools.chain(*row), boxed)
        self.assertEqual(map(list, pixels), map(list, flat))
    def testRGBtoRGBA(self):
        """asRGBA8() on colour type 2 source."""
        # Test for Issue 26 (googlecode)
        # Also test that png.Reader can take a "file-like" object.
        r = png.Reader(BytesIO(pngsuite.basn2c08))
        x,y,pixels,meta = r.asRGBA8()
        # Test the pixels at row 9 columns 0 and 1.
        row9 = list(pixels)[9]
        self.assertEqual(list(row9[0:8]),
                         [0xff, 0xdf, 0xff, 0xff, 0xff, 0xde, 0xff, 0xff])
    def testLtoRGBA(self):
        """asRGBA() on grey source."""
        # Test for Issue 60 (googlecode)
        r = png.Reader(bytes=pngsuite.basi0g08)
        x,y,pixels,meta = r.asRGBA()
        row9 = list(list(pixels)[9])
        self.assertEqual(row9[0:8],
          [222, 222, 222, 255, 221, 221, 221, 255])
    def testCtrns(self):
        "Test colour type 2 and tRNS chunk."
        # Test for Issue 25 (googlecode)
        r = png.Reader(bytes=pngsuite.tbrn2c08)
        x,y,pixels,meta = r.asRGBA8()
        # I just happen to know that the first pixel is transparent.
        # In particular it should be #7f7f7f00
        row0 = list(pixels)[0]
        self.assertEqual(tuple(row0[0:4]), (0x7f, 0x7f, 0x7f, 0x00))
    def testAdam7read(self):
        """Adam7 interlace reading.
        Specifically, test that for images in the PngSuite that
        have both an interlaced and straightlaced pair that both
        images from the pair produce the same array of pixels."""
        for candidate in pngsuite.png:
            if not candidate.startswith('basn'):
                continue
            candi = candidate.replace('n', 'i')
            if candi not in pngsuite.png:
                continue
            print 'adam7 read', candidate
            straight = png.Reader(bytes=pngsuite.png[candidate])
            adam7 = png.Reader(bytes=pngsuite.png[candi])
            # Just compare the pixels.  Ignore x,y (because they're
            # likely to be correct?); metadata is ignored because the
            # "interlace" member differs.  Lame.
            straight = straight.read()[2]
            adam7 = adam7.read()[2]
            self.assertEqual(map(list, straight), map(list, adam7))
    def testAdam7write(self):
        """Adam7 interlace writing.
        For each test image in the PngSuite, write an interlaced
        and a straightlaced version.  Decode both, and compare results.
        """
        # Not such a great test, because the only way we can check what
        # we have written is to read it back again.

        for name,bytes in pngsuite.png.iteritems():
            # Only certain colour types supported for this test.
            if name[3:5] not in ['n0', 'n2', 'n4', 'n6']:
                continue
            it = png.Reader(bytes=bytes)
            x,y,pixels,meta = it.read()
            pngi = topngbytes('adam7wn'+name+'.png', pixels,
              x=x, y=y, bitdepth=it.bitdepth,
              greyscale=it.greyscale, alpha=it.alpha,
              transparent=it.transparent,
              interlace=False)
            x,y,ps,meta = png.Reader(bytes=pngi).read()
            it = png.Reader(bytes=bytes)
            x,y,pixels,meta = it.read()
            pngs = topngbytes('adam7wi'+name+'.png', pixels,
              x=x, y=y, bitdepth=it.bitdepth,
              greyscale=it.greyscale, alpha=it.alpha,
              transparent=it.transparent,
              interlace=True)
            x,y,pi,meta = png.Reader(bytes=pngs).read()
            self.assertEqual(map(list, ps), map(list, pi))
    def testPGMin(self):
        """Test that the command line tool can read PGM files."""
        def do():
            return png._main(['testPGMin'])
        s = BytesIO()
        s.write(strtobytes('P5 2 2 3\n'))
        s.write(strtobytes('\x00\x01\x02\x03'))
        s.flush()
        s.seek(0)
        o = BytesIO()
        _redirect_io(s, o, do)
        r = png.Reader(bytes=o.getvalue())
        x,y,pixels,meta = r.read()
        self.assertTrue(r.greyscale)
        self.assertEqual(r.bitdepth, 2)
    def testPAMin(self):
        """Test that the command line tool can read PAM file."""
        def do():
            return png._main(['testPAMin'])
        s = BytesIO()
        s.write(strtobytes('P7\nWIDTH 3\nHEIGHT 1\nDEPTH 4\nMAXVAL 255\n'
                'TUPLTYPE RGB_ALPHA\nENDHDR\n'))
        # The pixels in flat row flat pixel format
        flat =  [255,0,0,255, 0,255,0,120, 0,0,255,30]
        asbytes = seqtobytes(flat)
        s.write(asbytes)
        s.flush()
        s.seek(0)
        o = BytesIO()
        _redirect_io(s, o, do)
        r = png.Reader(bytes=o.getvalue())
        x,y,pixels,meta = r.read()
        self.assertTrue(r.alpha)
        self.assertTrue(not r.greyscale)
        self.assertEqual(list(itertools.chain(*pixels)), flat)
    def testLA4(self):
        """Create an LA image with bitdepth 4."""
        bytes = topngbytes('la4.png', [[5, 12]], 1, 1,
          greyscale=True, alpha=True, bitdepth=4)
        sbit = png.Reader(bytes=bytes).chunk('sBIT')[1]
        self.assertEqual(sbit, strtobytes('\x04\x04'))
    def testPal(self):
        """Test that a palette PNG returns the palette in info."""
        r = png.Reader(bytes=pngsuite.basn3p04)
        x,y,pixels,info = r.read()
        self.assertEqual(x, 32)
        self.assertEqual(y, 32)
        self.assertTrue('palette' in info)
    def testPalWrite(self):
        """Test metadata for paletted PNG can be passed from one PNG
        to another."""
        r = png.Reader(bytes=pngsuite.basn3p04)
        x,y,pixels,info = r.read()
        w = png.Writer(**info)
        o = BytesIO()
        w.write(o, pixels)
        o.flush()
        o.seek(0)
        r = png.Reader(file=o)
        _,_,_,again_info = r.read()
        # Same palette
        self.assertEqual(again_info['palette'], info['palette'])
    def testPalExpand(self):
        """Test that bitdepth can be used to fiddle with pallete image."""
        r = png.Reader(bytes=pngsuite.basn3p04)
        x,y,pixels,info = r.read()
        pixels = [list(row) for row in pixels]
        info['bitdepth'] = 8
        w = png.Writer(**info)
        o = BytesIO()
        w.write(o, pixels)
        o.flush()
        o.seek(0)
        r = png.Reader(file=o)
        _,_,again_pixels,again_info = r.read()
        # Same pixels
        again_pixels = [list(row) for row in again_pixels]
        self.assertEqual(again_pixels, pixels)

    def testPNMsbit(self):
        """Test that PNM files can generates sBIT chunk."""
        def do():
            return png._main(['testPNMsbit'])
        s = BytesIO()
        s.write(strtobytes('P6 8 1 1\n'))
        for pixel in range(8):
            s.write(struct.pack('<I', (0x4081*pixel)&0x10101)[:3])
        s.flush()
        s.seek(0)
        o = BytesIO()
        _redirect_io(s, o, do)
        r = png.Reader(bytes=o.getvalue())
        sbit = r.chunk('sBIT')[1]
        self.assertEqual(sbit, strtobytes('\x01\x01\x01'))
    def testLtrns0(self):
        """Create greyscale image with tRNS chunk."""
        return self.helperLtrns(0)
    def testLtrns1(self):
        """Using 1-tuple for transparent arg."""
        return self.helperLtrns((0,))
    def helperLtrns(self, transparent):
        """Helper used by :meth:`testLtrns*`."""
        pixels = zip([0x00, 0x38, 0x4c, 0x54, 0x5c, 0x40, 0x38, 0x00])
        o = BytesIO()
        w = png.Writer(8, 8, greyscale=True, bitdepth=1, transparent=transparent)
        w.write_packed(o, pixels)
        r = png.Reader(bytes=o.getvalue())
        x,y,pixels,meta = r.asDirect()
        self.assertTrue(meta['alpha'])
        self.assertTrue(meta['greyscale'])
        self.assertEqual(meta['bitdepth'], 1)
    def testWinfo(self):
        """Test the dictionary returned by a `read` method can be used
        as args for :meth:`Writer`.
        """
        r = png.Reader(bytes=pngsuite.basn2c16)
        info = r.read()[3]
        w = png.Writer(**info)
    def testPackedIter(self):
        """Test iterator for row when using write_packed.

        Indicative for Issue 47 (googlecode).
        """
        w = png.Writer(16, 2, greyscale=True, alpha=False, bitdepth=1)
        o = BytesIO()
        w.write_packed(o, [itertools.chain([0x0a], [0xaa]),
                           itertools.chain([0x0f], [0xff])])
        r = png.Reader(bytes=o.getvalue())
        x,y,pixels,info = r.asDirect()
        pixels = list(pixels)
        self.assertEqual(len(pixels), 2)
        self.assertEqual(len(pixels[0]), 16)
    def testInterlacedArray(self):
        """Test that reading an interlaced PNG yields each row as an
        array."""
        r = png.Reader(bytes=pngsuite.basi0g08)
        list(r.read()[2])[0].tostring
    def testTrnsArray(self):
        """Test that reading a type 2 PNG with tRNS chunk yields each
        row as an array (using asDirect)."""
        r = png.Reader(bytes=pngsuite.tbrn2c08)
        list(r.asDirect()[2])[0].tostring

    # Invalid file format tests.  These construct various badly
    # formatted PNG files, then feed them into a Reader.  When
    # everything is working properly, we should get FormatError
    # exceptions raised.
    def testEmpty(self):
        """Test empty file."""

        r = png.Reader(bytes='')
        self.assertRaises(png.FormatError, r.asDirect)
    def testSigOnly(self):
        """Test file containing just signature bytes."""

        r = png.Reader(bytes=pngsuite.basi0g01[:8])
        self.assertRaises(png.FormatError, r.asDirect)
    def testChun(self):
        """
        Chunk doesn't have length and type.
        """
        r = png.Reader(bytes=pngsuite.basi0g01[:13])
        try:
            r.asDirect()
        except Exception as e:
            self.assertIsInstance(e, png.FormatError)
            self.assertIn('chunk length', str(e))
    def testChunkShort(self):
        """
        Chunk that is too short.
        """
        r = png.Reader(bytes=pngsuite.basi0g01[:21])
        try:
            r.asDirect()
        except Exception as e:
            self.assertIsInstance(e, png.ChunkError)
            self.assertIn('too short', str(e))
    def testNoChecksum(self):
        """
        Chunk that's too small to contain a checksum.
        """
        r = png.Reader(bytes=pngsuite.basi0g01[:29])
        try:
            r.asDirect()
        except Exception as e:
            self.assertIsInstance(e, png.ChunkError)
            self.assertIn('checksum', str(e))
    def testExtraPixels(self):
        """Test file that contains too many pixels."""

        def eachchunk(chunk):
            if chunk[0] != 'IDAT':
                return chunk
            data = zlib.decompress(chunk[1])
            data += strtobytes('\x00garbage')
            data = zlib.compress(data)
            chunk = (chunk[0], data)
            return chunk
        self.assertRaises(png.FormatError, self.helperFormat, eachchunk)
    def testNotEnoughPixels(self):
        def eachchunk(chunk):
            if chunk[0] != 'IDAT':
                return chunk
            # Remove last byte.
            data = zlib.decompress(chunk[1])
            data = data[:-1]
            data = zlib.compress(data)
            return (chunk[0], data)
        self.assertRaises(png.FormatError, self.helperFormat, eachchunk)
    def helperFormat(self, f):
        r = png.Reader(bytes=pngsuite.basn0g01)
        o = BytesIO()
        def newchunks():
            for chunk in r.chunks():
                yield f(chunk)
        png.write_chunks(o, newchunks())
        r = png.Reader(bytes=o.getvalue())
        return list(r.asDirect()[2])
    def testBadFilter(self):
        def eachchunk(chunk):
            if chunk[0] != 'IDAT':
                return chunk
            data = zlib.decompress(chunk[1])
            # Corrupt the first filter byte
            data = strtobytes('\x99') + data[1:]
            data = zlib.compress(data)
            return (chunk[0], data)
        self.assertRaises(png.FormatError, self.helperFormat, eachchunk)

    def testFlat(self):
        """Test read_flat."""
        import hashlib

        r = png.Reader(bytes=pngsuite.basn0g02)
        x,y,pixel,meta = r.read_flat()
        d = hashlib.md5(seqtobytes(pixel)).hexdigest()
        self.assertEqual(d, '255cd971ab8cd9e7275ff906e5041aa0')

    def testfromarray(self):
        img = png.from_array([[0, 0x33, 0x66], [0xff, 0xcc, 0x99]], 'L')
        img.save(BytesIO())
    def testfromarray3D(self):
        img = png.from_array(
            [[[0, 0, 0], [255, 0, 0]],
             [[255, 0, 0], [0, 0, 0]]], 'RGB')
        img.save(BytesIO())
    def testfromarrayL16(self):
        img = png.from_array(group(range(2**16), 256), 'L;16')
        img.save(BytesIO())
    def testfromarrayRGB(self):
        img = png.from_array([[0,0,0, 0,0,1, 0,1,0, 0,1,1],
                          [1,0,0, 1,0,1, 1,1,0, 1,1,1]], 'RGB;1')
        o = BytesIO()
        img.save(o)
    def testfromarrayIter(self):
        i = itertools.islice(itertools.count(10), 20)
        i = itertools.imap(lambda x: [x, x, x], i)
        img = png.from_array(i, 'RGB;5', dict(height=20))
        f = BytesIO()
        img.save(f)
    def testfromarrayWrong(self):
        try:
            png.from_array([[1]], 'gray')
        except png.Error:
            return
        assert 0, "Expected from_array() to raise png.Error exception"
    def testfromarrayShortMode(self):
        png.from_array([[0,1],[2,3]], 'L2').save(BytesIO())

    # numpy dependent tests.  These are skipped (with a message to
    # sys.stderr) if numpy cannot be imported.
    def testNumpyuint16(self):
        """numpy uint16."""

        try:
            import numpy
        except ImportError:
            print >>sys.stderr, "skipping numpy test"
            return

        rows = [map(numpy.uint16, range(0,0x10000,0x5555))]
        b = topngbytes('numpyuint16.png', rows, 4, 1,
            greyscale=True, alpha=False, bitdepth=16)
    def testNumpyuint8(self):
        """numpy uint8."""

        try:
            import numpy
        except ImportError:
            print >>sys.stderr, "skipping numpy test"
            return

        rows = [map(numpy.uint8, range(0,0x100,0x55))]
        b = topngbytes('numpyuint8.png', rows, 4, 1,
            greyscale=True, alpha=False, bitdepth=8)
    def testNumpybool(self):
        """numpy bool."""

        try:
            import numpy
        except ImportError:
            print >>sys.stderr, "skipping numpy test"
            return

        rows = [map(numpy.bool, [0,1])]
        b = topngbytes('numpybool.png', rows, 2, 1,
            greyscale=True, alpha=False, bitdepth=1)
    def testNumpyarray(self):
        """numpy array."""
        try:
            import numpy
        except ImportError:
            print >>sys.stderr, "skipping numpy test"
            return

        pixels = numpy.array([[0,0x5555],[0x5555,0xaaaa]], numpy.uint16)
        img = png.from_array(pixels, 'L')
        img.save(BytesIO())
    def testNumpyPalette(self):
        """numpy palette."""

        try:
            import numpy
        except ImportError:
            print >>sys.stderr, "skipping numpy test"
            return

        s = ['110010010011',
             '101011010100',
             '110010110101',
             '100010010011']

        s = map(lambda x: map(int, x), s)

        palette = [(0x55,0x55,0x55), (0xff,0x99,0x99)]
        pnp = numpy.array(palette) # creates a 2x3 array
        w = png.Writer(len(s[0]), len(s), palette=pnp, bitdepth=1)

    def paeth(self, x, a, b, c):
        p = a + b - c
        pa = abs(p - a)
        pb = abs(p - b)
        pc = abs(p - c)
        if pa <= pb and pa <= pc:
            pr = a
        elif pb <= pc:
            pr = b
        else:
            pr = c
        return x - pr

    # test filters and unfilters
    def testFilterScanlineFirstLine(self):
        fo = 3  # bytes per pixel
        line = [30, 31, 32, 230, 231, 232]
        out = png.filter_scanline(0, line, fo, None)  # none
        self.assertEqual(list(out), [0, 30, 31, 32, 230, 231, 232])
        out = png.filter_scanline(1, line, fo, None)  # sub
        self.assertEqual(list(out), [1, 30, 31, 32, 200, 200, 200])
        out = png.filter_scanline(2, line, fo, None)  # up
        self.assertEqual(list(out), [2, 30, 31, 32, 230, 231, 232])
        out = png.filter_scanline(3, line, fo, None)  # average
        self.assertEqual(list(out), [3, 30, 31, 32, 215, 216, 216])
        out = png.filter_scanline(4, line, fo, None)  # paeth
        self.assertEqual(list(out), [
            4, self.paeth(30, 0, 0, 0), self.paeth(31, 0, 0, 0),
            self.paeth(32, 0, 0, 0), self.paeth(230, 30, 0, 0),
            self.paeth(231, 31, 0, 0), self.paeth(232, 32, 0, 0)
            ])
    def testFilterScanline(self):
        prev = [20, 21, 22, 210, 211, 212]
        line = [30, 32, 34, 230, 233, 236]
        fo = 3
        out = png.filter_scanline(0, line, fo, prev)  # none
        self.assertEqual(list(out), [0, 30, 32, 34, 230, 233, 236])
        out = png.filter_scanline(1, line, fo, prev)  # sub
        self.assertEqual(list(out), [1, 30, 32, 34, 200, 201, 202])
        out = png.filter_scanline(2, line, fo, prev)  # up
        self.assertEqual(list(out), [2, 10, 11, 12, 20, 22, 24])
        out = png.filter_scanline(3, line, fo, prev)  # average
        self.assertEqual(list(out), [3, 20, 22, 23, 110, 112, 113])
        out = png.filter_scanline(4, line, fo, prev)  # paeth
        self.assertEqual(list(out), [
            4, self.paeth(30, 0, 20, 0), self.paeth(32, 0, 21, 0),
            self.paeth(34, 0, 22, 0), self.paeth(230, 30, 210, 20),
            self.paeth(233, 32, 211, 21), self.paeth(236, 34, 212, 22)
            ])
    def testUnfilterScanline(self):
        reader = png.Reader(bytes='')
        reader.psize = 3
        scanprev = array('B', [20, 21, 22, 210, 211, 212])
        scanline = array('B', [30, 32, 34, 230, 233, 236])
        def cp(a):
            return array('B', a)

        out = reader.undo_filter(0, cp(scanline), cp(scanprev))
        self.assertEqual(list(out), list(scanline))  # none
        out = reader.undo_filter(1, cp(scanline), cp(scanprev))
        self.assertEqual(list(out), [30, 32, 34, 4, 9, 14])  # sub
        out = reader.undo_filter(2, cp(scanline), cp(scanprev))
        self.assertEqual(list(out), [50, 53, 56, 184, 188, 192])  # up
        out = reader.undo_filter(3, cp(scanline), cp(scanprev))
        self.assertEqual(list(out), [40, 42, 45, 99, 103, 108])  # average
        out = reader.undo_filter(4, cp(scanline), cp(scanprev))
        self.assertEqual(list(out), [50, 53, 56, 184, 188, 192])  # paeth
    def testUnfilterScanlinePaeth(self):
        # This tests more edge cases in the paeth unfilter
        reader = png.Reader(bytes='')
        reader.psize = 3
        scanprev = array('B', [2, 0, 0, 0, 9, 11])
        scanline = array('B', [6, 10, 9, 100, 101, 102])

        out = reader.undo_filter(4, scanline, scanprev)
        self.assertEqual(list(out), [8, 10, 9, 108, 111, 113])  # paeth

    def testModifyRows(self):
        # Tests that the rows yielded by the pixels generator
        # can be safely modified.
        k = 'f02n0g08'
        r1 = png.Reader(bytes=pngsuite.png[k])
        r2 = png.Reader(bytes=pngsuite.png[k])
        _,_,pixels1,info1 = r1.asDirect()
        _,_,pixels2,info2 = r2.asDirect()
        for row1, row2 in itertools.izip(pixels1, pixels2):
            self.assertEqual(row1, row2)
            for i in range(len(row1)):
                row1[i] = 11117 % (i + 1)

    def testPNMWrite(self):
        o = BytesIO()
        w,h = 3,3
        pixels = [[0, 1, 2],
                  [3, 0, 1],
                  [2, 3, 0]]
        meta = dict(alpha=False, greyscale=True, bitdepth=2, planes=1)
        png.write_pnm(o, w, h, pixels, meta)

def group(s, n):
    # See http://www.python.org/doc/2.6/library/functions.html#zip
    return zip(*[iter(s)]*n)

if __name__ == '__main__':
    unittest.main(__name__)



#cython: boundscheck=False
#cython: wraparound=False

from libc.stdlib cimport abs as c_abs

cimport cpython.array


# TODO: I don't know how can I not return any value (void doesn't work)
cpdef int undo_filter_sub(int filter_unit, unsigned char[:] scanline,
                          unsigned char[:] previous, unsigned char[:] result) nogil:
    """Undo sub filter."""

    cdef int l = result.shape[0]
    cdef int ai = 0
    cdef unsigned char x, a

    # Loops starts at index fu.  Observe that the initial part
    # of the result is already filled in correctly with
    # scanline.
    for i in range(filter_unit, l):
        x = scanline[i]
        a = result[ai]
        result[i] = (x + a) & 0xff
        ai += 1
    return 0


cpdef int undo_filter_up(int filter_unit, unsigned char[:] scanline,
                         unsigned char[:] previous, unsigned char[:] result) nogil:
    """Undo up filter."""

    cdef int i
    cdef int l = result.shape[0]
    cdef unsigned char x, b

    for i in range(l):
        x = scanline[i]
        b = previous[i]
        result[i] = (x + b) & 0xff
    return 0


cpdef int undo_filter_average(int filter_unit, unsigned char[:] scanline,
                              unsigned char[:] previous, unsigned char[:] result) nogil:
    """Undo up filter."""

    cdef int i, ai
    cdef int l = result.shape[0]
    cdef unsigned char x, a, b

    ai = -filter_unit
    for i in range(l):
        x = scanline[i]
        if ai < 0:
            a = 0
        else:
            a = result[ai]
        b = previous[i]
        result[i] = (x + ((a + b) >> 1)) & 0xff
        ai += 1
    return 0


cpdef int undo_filter_paeth(int filter_unit, unsigned char[:] scanline,
                            unsigned char[:] previous, unsigned char[:] result) nogil:
    """Undo Paeth filter."""

    # Also used for ci.
    cdef int ai = -filter_unit
    cdef int l = result.shape[0]
    cdef int i, pa, pb, pc, p
    cdef unsigned char x, a, b, c, pr

    for i in range(l):
        x = scanline[i]
        if ai < 0:
            a = c = 0
        else:
            a = result[ai]
            c = previous[ai]
        b = previous[i]
        p = a + b - c
        pa = c_abs(p - a)
        pb = c_abs(p - b)
        pc = c_abs(p - c)
        if pa <= pb and pa <= pc:
            pr = a
        elif pb <= pc:
            pr = b
        else:
            pr = c
        result[i] = (x + pr) & 0xff
        ai += 1
    return 0


cpdef int convert_rgb_to_rgba(unsigned char[:] row, unsigned char[:] result) nogil:
    cdef int i, l, j, k
    l = min(row.shape[0] / 3, result.shape[0] / 4)
    for i in range(l):
        j = i * 3
        k = i * 4
        result[k] = row[j]
        result[k + 1] = row[j + 1]
        result[k + 2] = row[j + 2]
    return 0


cpdef int convert_l_to_rgba(unsigned char[:] row, unsigned char[:] result) nogil:
    cdef int i, l, j, k, lum
    l = min(row.shape[0], result.shape[0] / 4)
    for i in range(l):
        j = i
        k = i * 4
        lum = row[j]
        result[k] = lum
        result[k + 1] = lum
        result[k + 2] = lum
    return 0


cpdef int convert_la_to_rgba(unsigned char[:] row, unsigned char[:] result) nogil:
    cdef int i, l, j, k, lum
    l = min(row.shape[0] / 2, result.shape[0] / 4)
    for i in range(l):
        j = i * 2
        k = i * 4
        lum = row[j]
        result[k] = lum
        result[k + 1] = lum
        result[k + 2] = lum
        result[k + 3] = row[j + 1]
    return 0



#!/usr/bin/env python
# $Rev$
# $URL$

# Imported from //depot/prj/plan9topam/master/code/plan9topam.py#4 on
# 2009-06-15.

"""Command line tool to convert from Plan 9 image format to PNG format.

Plan 9 image format description:
http://plan9.bell-labs.com/magic/man2html/6/image
"""

# http://www.python.org/doc/2.3.5/lib/module-itertools.html
import itertools
# http://www.python.org/doc/2.3.5/lib/module-re.html
import re
# http://www.python.org/doc/2.3.5/lib/module-sys.html
import sys

class Error(Exception):
    """Some sort of Plan 9 image error."""

def block(s, n):
    # See http://www.python.org/doc/2.6.2/library/functions.html#zip
    return zip(*[iter(s)]*n)

def convert(f, output=sys.stdout) :
  """Convert Plan 9 file to PNG format.  Works with either uncompressed
  or compressed files.
  """

  r = f.read(11)
  if r == 'compressed\n' :
    png(output, *decompress(f))
  else :
    png(output, *glue(f, r))


def glue(f, r) :
  """Return (metadata, stream) pair where `r` is the initial portion of
  the metadata that has already been read from the stream `f`.
  """

  r = r + f.read(60-len(r))
  return (r, f)

def meta(r) :
  """Convert 60 character string `r`, the metadata from an image file.
  Returns a 5-tuple (*chan*,*minx*,*miny*,*limx*,*limy*).  5-tuples may
  settle into lists in transit.
  
  As per http://plan9.bell-labs.com/magic/man2html/6/image the metadata
  comprises 5 words separated by blanks.  As it happens each word starts
  at an index that is a multiple of 12, but this routine does not care
  about that."""

  r = r.split()
  # :todo: raise FormatError
  assert len(r) == 5
  r = [r[0]] + map(int, r[1:])
  return r

def bitdepthof(pixel) :
    """Return the bitdepth for a Plan9 pixel format string."""

    maxd = 0
    for c in re.findall(r'[a-z]\d*', pixel) :
        if c[0] != 'x':
            maxd = max(maxd, int(c[1:]))
    return maxd

def maxvalof(pixel):
  """Return the netpbm MAXVAL for a Plan9 pixel format string."""

  bitdepth = bitdepthof(pixel)
  return (2**bitdepth)-1

def pixmeta(metadata, f) :
    """Convert (uncompressed) Plan 9 image file to pair of (*metadata*,
    *pixels*).  This is intended to be used by PyPNG format.  *metadata*
    is the metadata returned in a dictionary, *pixels* is an iterator that
    yields each row in boxed row flat pixel format.

    `f`, the input file, should be cued up to the start of the image data.
    """

    chan,minx,miny,limx,limy = metadata
    rows = limy - miny
    width = limx - minx
    nchans = len(re.findall('[a-wyz]', chan))
    alpha = 'a' in chan
    # Iverson's convention for the win!
    ncolour = nchans - alpha
    greyscale = ncolour == 1
    bitdepth = bitdepthof(chan)
    maxval = 2**bitdepth - 1
    # PNG style metadata
    meta=dict(size=(width,rows), bitdepth=bitdepthof(chan),
      greyscale=greyscale, alpha=alpha, planes=nchans)

    return itertools.imap(lambda x: itertools.chain(*x),
      block(unpack(f, rows, width, chan, maxval), width)), meta

def png(out, metadata, f):
    """Convert to PNG format.  `metadata` should be a Plan9 5-tuple; `f`
    the input file (see :meth:`pixmeta`).
    """

    import png

    pixels,meta = pixmeta(metadata, f)
    p = png.Writer(**meta)
    p.write(out, pixels)

def spam():
  """Not really spam, but old PAM code, which is in limbo."""

  if nchans == 3 or nchans == 1 :
    # PGM (P5) or PPM (P6) format.
    output.write('P%d\n%d %d %d\n' % (5+(nchans==3), width, rows, maxval))
  else :
    # PAM format.
    output.write("""P7
WIDTH %d
HEIGHT %d
DEPTH %d
MAXVAL %d
""" % (width, rows, nchans, maxval))

def unpack(f, rows, width, pixel, maxval) :
  """Unpack `f` into pixels.  Assumes the pixel format is such that the depth
  is either a multiple or a divisor of 8.
  `f` is assumed to be an iterator that returns blocks of input such
  that each block contains a whole number of pixels.  An iterator is
  returned that yields each pixel as an n-tuple.  `pixel` describes the
  pixel format using the Plan9 syntax ("k8", "r8g8b8", and so on).
  """

  def mask(w) :
    """An integer, to be used as a mask, with bottom `w` bits set to 1."""

    return (1 << w)-1

  def deblock(f, depth, width) :
    """A "packer" used to convert multiple bytes into single pixels.
    `depth` is the pixel depth in bits (>= 8), `width` is the row width in
    pixels.
    """

    w = depth // 8
    i = 0
    for block in f :
      for i in range(len(block)//w) :
        p = block[w*i:w*(i+1)]
        i += w
        # Convert p to little-endian integer, x
        x = 0
        s = 1 # scale
        for j in p :
          x += s * ord(j)
          s <<= 8
        yield x

  def bitfunge(f, depth, width) :
    """A "packer" used to convert single bytes into multiple pixels.
    Depth is the pixel depth (< 8), width is the row width in pixels.
    """

    for block in f :
      col = 0
      for i in block :
        x = ord(i)
        for j in range(8/depth) :
          yield x >> (8 - depth)
          col += 1
          if col == width :
            # A row-end forces a new byte even if we haven't consumed
            # all of the current byte.  Effectively rows are bit-padded
            # to make a whole number of bytes.
            col = 0
            break
          x <<= depth

  # number of bits in each channel
  chan = map(int, re.findall(r'\d+', pixel))
  # type of each channel
  type = re.findall('[a-z]', pixel)

  depth = sum(chan)

  # According to the value of depth pick a "packer" that either gathers
  # multiple bytes into a single pixel (for depth >= 8) or split bytes
  # into several pixels (for depth < 8)
  if depth >= 8 :
    # 
    assert depth % 8 == 0
    packer = deblock
  else :
    assert 8 % depth == 0
    packer = bitfunge

  for x in packer(f, depth, width) :
    # x is the pixel as an unsigned integer
    o = []
    # This is a bit yucky.  Extract each channel from the _most_
    # significant part of x.
    for j in range(len(chan)) :
      v = (x >> (depth - chan[j])) & mask(chan[j])
      x <<= chan[j]
      if type[j] != 'x' :
        # scale to maxval
        v = v * float(maxval) / mask(chan[j])
        v = int(v+0.5)
        o.append(v)
    yield o


def decompress(f) :
  """Decompress a Plan 9 image file.  Assumes f is already cued past the
  initial 'compressed\n' string.
  """

  r = meta(f.read(60))
  return r, decomprest(f, r[4])


def decomprest(f, rows) :
  """Iterator that decompresses the rest of a file once the metadata
  have been consumed."""

  row = 0
  while row < rows :
    row,o = deblock(f)
    yield o


def deblock(f) :
  """Decompress a single block from a compressed Plan 9 image file.
  Each block starts with 2 decimal strings of 12 bytes each.  Yields a
  sequence of (row, data) pairs where row is the total number of rows
  processed according to the file format and data is the decompressed
  data for a set of rows."""

  row = int(f.read(12))
  size = int(f.read(12))
  if not (0 <= size <= 6000) :
    raise Error('block has invalid size; not a Plan 9 image file?')

  # Since each block is at most 6000 bytes we may as well read it all in
  # one go.
  d = f.read(size)
  i = 0
  o = []

  while i < size :
    x = ord(d[i])
    i += 1
    if x & 0x80 :
      x = (x & 0x7f) + 1
      lit = d[i:i+x]
      i += x
      o.extend(lit)
      continue
    # x's high-order bit is 0
    l = (x >> 2) + 3
    # Offset is made from bottom 2 bits of x and all 8 bits of next
    # byte.  http://plan9.bell-labs.com/magic/man2html/6/image doesn't
    # say whether x's 2 bits are most significant or least significant.
    # But it is clear from inspecting a random file,
    # http://plan9.bell-labs.com/sources/plan9/sys/games/lib/sokoban/images/cargo.bit
    # that x's 2 bit are most significant.
    # 
    offset = (x & 3) << 8
    offset |= ord(d[i])
    i += 1
    # Note: complement operator neatly maps (0 to 1023) to (-1 to
    # -1024).  Adding len(o) gives a (non-negative) offset into o from
    # which to start indexing.
    offset = ~offset + len(o)
    if offset < 0 :
      raise Error('byte offset indexes off the begininning of the output buffer; not a Plan 9 image file?')
    for j in range(l) :
      o.append(o[offset+j])
  return row,''.join(o)

def main(argv=None) :
  if argv is None :
    argv = sys.argv
  if len(sys.argv) <= 1 :
    return convert(sys.stdin)
  else :
    return convert(open(argv[1], 'rb'))

if __name__ == '__main__' :
  sys.exit(main())



# PyPNG setup.py
# This is the setup.py script used by distutils.

# You can install the png module into your Python distribution with:
# python setup.py install
# You can also do other standard distutil type things, but you can refer
# to the distutil documentation for that.

# This script is also imported as a module by the Sphinx conf.py script
# in the man directory, so that this file forms a single source for
# metadata.

# http://docs.python.org/release/2.4.4/lib/module-sys.html
import sys

def get_version():
    from os.path import dirname, join
    for line in open(join(dirname(__file__), "code", "png.py")):
        if '__version__' in line:
            version = line.split('"')[1]
            break
    return version

conf = dict(
    name='pypng',
    version=get_version(),
    description='Pure Python PNG image encoder/decoder',
    long_description="""
PyPNG allows PNG image files to be read and written using pure Python.

It's available from github.com
https://github.com/drj11/pypng

Documentation is kindly hosted by PyPI
http://pythonhosted.org/pypng/
(and also available in the download tarball).
""",
    author='David Jones',
    author_email='drj@pobox.com',
    url='https://github.com/drj11/pypng',
    package_dir={'':'code'},
    py_modules=['png'],
    classifiers=[
      'Topic :: Multimedia :: Graphics',
      'Topic :: Software Development :: Libraries :: Python Modules',
      'Programming Language :: Python',
      'Programming Language :: Python :: 2.3',
      'Programming Language :: Python :: 3',
      'License :: OSI Approved :: MIT License',
      'Operating System :: OS Independent',
      ],
    )
conf['download_url'] = \
  'https://github.com/drj11/pypng/archive/%(name)s-%(version)s.tar.gz' % conf

def prepare3():
    """Prepare files for installing on Python 3.  If you have
    distribute for Python 3, then we don't need to run this.
    """
    import os

    try:
        os.mkdir('code3')
    except OSError:
        pass
    # Note: -W was added in 2.7.3.
    from lib2to3.main import main
    main("lib2to3.fixes", ["-w", "-W", "-n", "-o", "code3",
                           conf['package_dir']['']])

    conf['package_dir'] = {'':'code3'}
      
if __name__ == '__main__':
    try:
        # http://peak.telecommunity.com/DevCenter/setuptools#basic-use
        from setuptools import setup
        # distribute is probably installed, so use_2to3 should work
        conf['use_2to3'] = True
    except ImportError:
        # http://docs.python.org/release/2.4.4/dist/setup-script.html
        from distutils.core import setup
        if sys.version_info >= (3,):
            prepare3()
    setup(**conf)
